import numpy as np
import numpy.linalg as la
import scipy.sparse as sparse
import sys

from .Mesh import Mesh

class Engine:
    def __init__(self, mesh: Mesh, order: int, formule: int):
        """Initialize the engine and assembles the support matrices 

        :param mesh: Instance of the mesh.
        :type mesh: FEM_3D.Mesh
        :param order: Order of the elements. Only order 1 is available.
        :type order: int
        :param formule: Formule for the numerical integration. Only 1 is available.
        :type formule: int
        """
        self.order = order
        self.formule = formule
        self.mesh = mesh
        self.dof,self.numDof,self.u,self.dxu,self.dyu,self.dzu,self.omega = self.initializeMatrices('Lagrange')
        
    def IntegrationPointsHat(self, formule: int):
        """Return the integration points of the reference tetrahedra

        :param formule: Formule for the integration. Only 1 is available.
        :type formule: int
        :return: (x,y,z,omega) ; x, y and z being the coordinates at the reference domain. Omega the weight of integration.
        :rtype: Tuple of arrays
        """
        xhat = []
        yhat = []
        zhat = []
        omega = []
        if (formule==1):
            a = (5-np.sqrt(5))/20
            b = (5+3*np.sqrt(5))/20
            xhat = [a,a,a,b]
            yhat = [a,a,b,a]
            zhat = [a,b,a,a]
            omega = [1/24,1/24,1/24,1/24]
        elif (formule==2):
            xhat = [1/4,1/2,1/6,1/6,1/6]
            yhat = [1/4,1/6,1/6,1/6,1/2]
            zhat = [1/4,1/6,1/6,1/2,1/6]
            omega = [-.8/6,.45/6,.45/6,.45/6,.45/6]
        return (np.array(xhat), np.array(yhat), np.array(zhat), np.array(omega))

    def initializeMatrices(self, typeEF: str):
        """Generate the matrices with the values of the base functions, the dx, dy, dz and the weights for integration.

        :param typeEF: Type of inteporlation. Only 'Lagrange'.
        :type typeEF: str
        :return: (dof,numDof,u,dxu,dyu,dzu,omega);
        dof is the position of the degrees of freedom
        numDof is the dofs that composes each element
        u is the value of the interpolation function for each integration point
        dxu is the derivative in x
        dyu is the derivative in y
        dzu is the derivative in z
        omega is the quadrature weight for each integration point
        :rtype: (Array n x 3, Array e x 3, Matrix i x e, Matrix i x e, Matrix i x e, Array i); 
        n being the number of degrees of freedom; 
        e being the number of elements;
        i being the number of integration points
        """
        mesh,order,formule = self.mesh, self.order, self.formule
        xhat, yhat, zhat, omegaloc = self.IntegrationPointsHat(formule)

        if (typeEF=='Lagrange'):
            phi, dxphi, dyphi, dzphi = self.LagrangeFunctions(xhat, yhat, zhat, order)

            if (order==1 or order==2): #GMSH generate the nodes
                dof = mesh.vertices
                numDof = np.array(mesh.tetrahedrons,dtype=np.int32)

            mesh.calcJacobians()
            detJ = mesh.determinants
            J_inv = mesh.inverses
                
            dxhat_dx = J_inv[:,0,0]
            dxhat_dy = J_inv[:,0,1]
            dxhat_dz = J_inv[:,0,2]
            dyhat_dx = J_inv[:,1,0]
            dyhat_dy = J_inv[:,1,1]
            dyhat_dz = J_inv[:,1,2]
            dzhat_dx = J_inv[:,2,0]
            dzhat_dy = J_inv[:,2,1]
            dzhat_dz = J_inv[:,2,2]

            N_dof = len(dof)
            N_int = len(mesh.tetrahedrons)*len(xhat)
            omega = np.zeros(N_int)

            #The sparse matrices will have for each line a integration point and each
            #column a dof. The number of integration points is the number of elements
            #times the number of points per element. So, for each point of integration
            #we have a certain amount of base functions to evaluate, the amount of 
            #dof per element.
            #Structure for sparse Matrix
            _i = np.zeros(N_int*len(phi[0]))
            _j = np.zeros(N_int*len(phi[0]))
            _v_u = np.zeros(N_int*len(phi[0]))
            _v_dxu = np.zeros(N_int*len(phi[0]))
            _v_dyu = np.zeros(N_int*len(phi[0]))
            _v_dzu = np.zeros(N_int*len(phi[0]))

            for iloc in range(0,len(xhat)): #Points of Integration
                iglob = np.arange(iloc,N_int,len(xhat),dtype=np.int64)
                omega[iglob] = omegaloc[iloc] * detJ
                for jloc in range(0,len(phi[0])): #Base functions
                    _idx = jloc*N_int + iloc*len(iglob) + np.arange(0,len(iglob)) #Global stacked index for the sparse matrix
                    jglob = numDof[:,jloc]
                    _i[_idx] = iglob
                    _j[_idx] = jglob
                    _v_u[_idx] = np.ones(len(iglob))*phi[iloc,jloc]
                    _v_dxu[_idx] = (dxphi[iloc,jloc]*dxhat_dx + dyphi[iloc,jloc]*dyhat_dx + dzphi[iloc,jloc]*dzhat_dx)
                    _v_dyu[_idx] = (dxphi[iloc,jloc]*dxhat_dy + dyphi[iloc,jloc]*dyhat_dy + dzphi[iloc,jloc]*dzhat_dy)
                    _v_dzu[_idx] = (dxphi[iloc,jloc]*dxhat_dz + dyphi[iloc,jloc]*dyhat_dz + dzphi[iloc,jloc]*dzhat_dz)
            
            
            #Generate the matrices from the row/col indexes and values
            u = sparse.coo_matrix((_v_u, (_i, _j)),(N_int,N_dof))
            dxu = sparse.coo_matrix((_v_dxu, (_i, _j)),(N_int,N_dof))
            dyu = sparse.coo_matrix((_v_dyu, (_i, _j)),(N_int,N_dof))
            dzu = sparse.coo_matrix((_v_dzu, (_i, _j)),(N_int,N_dof))
            u = u.tocsc() 
            dxu = dxu.tocsc()
            dyu = dyu.tocsc()
            dzu = dzu.tocsc()

        return np.array(dof),np.array(numDof),u,dxu,dyu,dzu,omega

    def LagrangeFunctions(self, x, y, z, order: int):
        """ Get the values and of the base functions and its derivatives at the tetrahedra of reference.

        :param x: Array of x positions in the domain of reference
        :type x: array(float)
        :param y: Array of y positions in the domain of reference
        :type y: array(float)
        :param z: Array of z positions in the domain of reference
        :type z: array(float)
        :param order: Order of the Lagrange function. Only 1 is available.
        :type order: int
        :raises Exception: If the Lagrange function order are no implemented
        :return: (value,x derivative,y derivative,z derivative)
        :rtype: (array(float),array(float),array(float),array(float))
        """
        phi = []
        dxphi = []
        dyphi = []
        dzphi = []
        if (order==1): #Ordering (0,0,0), (1,0,0), (0,1,0) et (0,0,1)
            for i in range(0,len(x)):
                phi.append([1-x[i]-y[i]-z[i],x[i],y[i],z[i]])
                dxphi.append([-1.,1.,0.,0])
                dyphi.append([-1.,0.,1.,0])
                dzphi.append([-1.,0.,0.,1])
        else: 
            raise Exception('Lagrande functions of order ' + str(order) + ' not implemented.')
        return np.array(phi), np.array(dxphi), np.array(dyphi), np.array(dzphi)
    
    def M_Matrix(self):
        """ Generate the mass matrix from the support matrices.

        :return: The mass matrix.
        :rtype: CSC Matrix
        """
        u,omega  = self.u,self.omega
        M = u.transpose().dot(u.transpose().multiply(omega).transpose())
        M = M.tocsc()

        #Garantee int64 to UMFPACK
        if 'scikits.umfpack' in sys.modules:
            M.indices = M.indices.astype(np.int64)
            M.indptr = M.indptr.astype(np.int64)
        return M

    def K_Matrix(self):
        """ Generate the stiffness matrix from the support matrices.

        :return: The stiffness matrix.
        :rtype: CSC Matrix
        """
        dxu,dyu,dzu,omega  = self.dxu,self.dyu,self.dzu,self.omega
        G1 = dxu.transpose().dot(dxu.transpose().multiply(omega).transpose())
        G2 = dyu.transpose().dot(dyu.transpose().multiply(omega).transpose())
        G3 = dzu.transpose().dot(dzu.transpose().multiply(omega).transpose())
        K = G1+G2+G3
        K = K.tocsc()

        #Garantee int64 to UMFPACK
        K.indices = K.indices.astype(np.int64)
        K.indptr = K.indptr.astype(np.int64)
        return K

    def F_Matrix(self, f):
        """Generate the forcing matrix.

        :param f: The source-term with argument time_index that returns an array with the values for each dof.
        :type f: function(int): Array
        :return: A function that returns the F matrix for each time index
        :rtype: function(int): Array
        """
        u,omega  = self.u,self.omega
        _ret = lambda time_index: u.multiply(f(time_index)).transpose().dot(omega)
        return _ret

    def getValueByPosition(self, dofValues, position):
        """Given the values of the degrees of freedom, it interpolates the lagrange functions to give the estimated value. Only works for Lagrange P1 elements.

        :param dofValues: The values at all degrees of freedom
        :type dofValues: Array
        :param position: Position to be evaluated using intepolation
        :type position: Array(3)
        :return: The interpolated value at the position.
        :rtype: float
        """
        etags = []
        coords_u = []
        coords_v = []
        coords_w = []
        for pos in position:
            x = pos[0]
            y = pos[1]
            z = pos[2]
            etag,etype,nodeTags,u,v,w = self.mesh.model.mesh.getElementByCoordinates(x,y,z,3)
            etags.append(etag)
            coords_u.append(u)
            coords_v.append(v)
            coords_w.append(w)
        tetraIndex = [np.where(self.mesh.tetraTags==etag)[0][0] for etag in etags]
        nodeIndex = self.mesh.tetrahedrons[tetraIndex]
        lFunction = self.LagrangeFunctions(coords_u,coords_v,coords_w,1)[0][0]
        return np.dot(lFunction, np.transpose(dofValues[:,nodeIndex],(0,2,1)))

    


import scipy.sparse.linalg as sla
from scipy.sparse.csgraph import reverse_cuthill_mckee
import sys


class umfLA(sla.LinearOperator):
    """ Linear operator which can use UMFPack to solve a linear system. """
    def __init__(self, M):
        self.solve = sla.factorized(M)
        self.shape = M.shape

    def _matvec(self, x):
        return self.solve(x)

class Solver():
    def __init__(self,engine):
        self.engine = engine
        if 'scikits.umfpack' in sys.modules:
            sla.use_solver(useUmfpack=True)

    def eig(self,A,M,k,sigma,which):
        """ Solve the eigenproblem A*x = a*M*x around sigma returning k eigenvalues/eigenvectors. """
        mi = umfLA(A)
        return sla.eigsh(A,k,M,sigma=sigma,which=which,OPinv=mi)

    def solve(self,f):
        """ Solves (K+M)x = F """
        A = self.engine.K_Matrix() + self.engine.M_Matrix()
        F = self.engine.F_Matrix(f)
        
        return sla.spsolve(A,F)

    def solve(self,A,F):
        """ Solves (A)x = F """
        
        return sla.spsolve(A,F)


import gmsh
import numpy as np
import matplotlib.pyplot as plt

class Mesh:
    @staticmethod
    def MeshSquare(name,Lx,Ly,h,order=1):
        """Create and return an instance of a square mesh.

        :param name: Name of the mesh.
        :type name: str
        :param Lx: x length of the square
        :type Lx: float
        :param Ly: x length of the square
        :type Ly: float
        :param h: element size
        :type h: float
        :param order: Order of the element [1 or 2], defaults to 1
        :type order: int, optional
        :return: The instance of the Mesh.
        :rtype: FEM_2D.Mesh
        """
        m = Mesh(name)
        elements = m.createSurface([[0,0,0],[Lx,0,0],[Lx,Ly,0],[0,Ly,0]],h=h)
        m.fac.addPlaneSurface([elements[-1]])
        m.generate(order=order)
        return m

    @staticmethod
    def MeshByTriangles(name,nodeTags,triangles,vertices):
        """Create and return a mesh instance given the triangles and its vertices. 
        The input must be ordered and numerated correctly.

        :param name: Name of the mesh
        :type name: str
        :param nodeTags: Identifier for each node. Used to map between different domains.
        :type nodeTags: Array
        :param triangles: The index of the vertices for each triangle.
        :type triangles: Array t x 3
        :param vertices: The coordinates of the vertices
        :type vertices: Array v x 3
        :return: The instance of the mesh.
        :rtype: FEM_2D.Mesh
        """
        m = Mesh(name)
        m.nodeTags = nodeTags
        m.vertices = vertices
        m.triangles = triangles
        return m

    def __init__(self, name):
        """Initialize GMSH and the mesh attributes

        :param name: Mesh name
        :type name: str
        """
        gmsh.initialize()
        gmsh.model.add(name)
        self.name = name
        self.model = gmsh.model
        self.fac = self.model.geo
        self.pos = gmsh.view
        self.vertices = np.array([])
        self.triangles = np.array([])

    def calcAreasRatio3D(self):
        """ Calculate the area of each triangle(element) for triangles in 3 dimensions. 
        Used for assembling the mass matrix for robin boundary conditions at surfaces.

        :return: The areas ratio between the triangle and the triangle of reference
        :rtype: Array(float)
        """
        K = self.vertices[self.triangles]

        #Structure of K:
        #x = Tk(x') = K(x'); for x' 1(0,0,1), 2(1,0,0), 3(0,1,0):

        #Only need the edges (first 3 vertices) to get the transformation
        K = K[:,0:3,:]
        K = np.moveaxis(K,2,1)
        
        #Area 
        X1 = K[:,:,1] - K[:,:,0]
        X2 = K[:,:,2] - K[:,:,0]
        x1=X1[:,0];y1=X1[:,1];z1=X1[:,2]
        x2=X2[:,0];y2=X2[:,1];z2=X2[:,2]
        A_paralelo = np.sqrt((y1*z2-y2*z1)**2 +  (x1*z2-x2*z1)**2 + (x1*y2 - x2*y1)**2) #TODO: Computational cost can be reduced?
        return 2 * A_paralelo/2

    def calcJacobians(self):
        """Pre-calculate the determinants and inverse of the jacobian for all triangles.
        Store the values at .determinants and .inverses
        """
        K = self.vertices[self.triangles]

        #Structure of K:
        #x = Tk(x') = K(x'); for x' 1(0,0,1), 2(1,0,0), 3(0,1,0):

        #Transformation Matrix
        xa = K[:,0,0];ya = K[:,0,1]
        xb = K[:,1,0];yb = K[:,1,1]
        xc = K[:,2,0];yc = K[:,2,1]
        a = (xb-xa)
        b = (xc-xa)
        c = (yb-ya)
        d = (yc-ya)
        #Inversion of 2x2 matrices
        self.determinants = a*d - b*c
        self.inverses = np.stack(1/self.determinants * [d,-b,-c,a],axis=1).reshape(-1,2,2)
        
    def createSurface(self, orderedVertices, h):
        """Uses the built-in GMSH mesh factory to construct a surface.

        :param orderedVertices: The coordinates of the surface vertices in order.
        :type orderedVertices: Array v x 3
        :param h: Elements size
        :type h: float
        :return: (p,l,cl) 
        p is the points tags
        l is the lines tags
        cl is the curve loops tag
        :rtype: (Array(int),Array(int),int)
        """
        p = []
        l = []
        for n in orderedVertices:
            p.append(self.fac.addPoint(n[0],n[1],n[2],h))

        for n in p:
            if (n < p[-1]):
                l.append(self.fac.addLine(n,n+1))
            else:
                l.append(self.fac.addLine(n,p[0]))

        cl = self.fac.addCurveLoop(l)

        return (p,l,cl)

    def createCircle(self, x, y, z, radius, h=0.03, addInterior = True):
        """Add a circle to the mesh. If addInterior = True, the interior of the circle will be meshed.

        :param x: The center coordinate x
        :type x: float
        :param y: The center coordinate y
        :type y: float
        :param z: The center coordinate z
        :type z: float
        :param radius: Circle radius
        :type radius: float
        :param h: Elements size, defaults to 0.03
        :type h: float, optional
        :param addInterior: If the elements will be meshed at the interior of the circle, defaults to True
        :type addInterior: bool, optional
        :return: (p,ca,cl,s)
        p are the points tags
        ca are the circle-arc tags
        cl is the curve loop tag
        s is the surface tag
        :rtype: (Array,Array,int,int)
        """
        p1 = self.fac.addPoint(x,y+radius,z,h)
        p2 = self.fac.addPoint(x,y,z,h)
        p3 = self.fac.addPoint(x,y-radius,z,h)
        ca1=self.fac.addCircleArc(p1,p2,p3)
        ca2=self.fac.addCircleArc(p3,p2,p1)

        cl = self.fac.addCurveLoop([ca1,ca2])

        s = self.fac.addPlaneSurface([cl]) if addInterior else 0

        return ([p1,p2,p3],[ca1,ca2],cl,s)

    def generate(self,order=1):
        """Generate the triangle elements of the mesh with GMSH

        :param order: Order of the elements [1 or 2], defaults to 1
        :type order: int, optional
        """
        self.fac.synchronize()
        self.model.mesh.generate(2)
        self.model.mesh.setOrder(order)
        self.model.mesh.removeDuplicateNodes()
        self.__fillVtxTri(order)
        
    def readMsh(self,file):
        """Read a mshFile with GMSH. The elements must be triangles of order 1.

        :param file: File path/name
        :type file: str
        """
        gmsh.open(file)
        self.fac.synchronize()
        self.__fillVtxTri(1)

    def __fillVtxTri(self,order):
        """Load the mesh infos(nodes and elements) from GMSH.

        :param order: Order of the elements [1 or 2]
        :type order: int
        """
        if order==1:
            tnodes = self.model.mesh.getNodesByElementType(2)
            unique, unique_indexes, unique_inverse = np.unique(tnodes[0],return_index=True,return_inverse=True)
            self.nodeTags = unique
            self.vertices = tnodes[1].reshape(-1,3)[unique_indexes]
            self.triangles = unique_inverse.reshape(-1,3)
        elif order==2:
            tnodes = self.model.mesh.getNodesByElementType(9)
            unique, unique_indexes, unique_inverse = np.unique(tnodes[0],return_index=True,return_inverse=True)
            self.nodeTags = unique
            self.vertices = tnodes[1].reshape(-1,3)[unique_indexes]
            self.triangles = unique_inverse.reshape(-1,6)

        self.vertice_group = np.zeros(len(self.vertices))
        for n in self.model.getPhysicalGroups():
            self.vertice_group[self.model.mesh.getNodesForPhysicalGroup(n[0],n[1])[0]-1] = n[1]

    def FLTKRun(self):
        """Opens the GMSH interface
        """
        gmsh.fltk.run()


import numpy as np
import gmsh
from . import FEM_2D as fem2d
from . import FEM_3D as fem3d

class Boundary:
    @staticmethod
    def Apply_Dirichlet(engine, physical_group_id, M_full, C_full, K_full, f=None, g=None): #TODO: what do you expect for g(time and dof or only time)?
        """Apply the dirichlet boundary condition given the matrices, 
        the physical group and f,g being functions in time index returning vectors with the values at the dofs.

        :param engine: The fem engine.
        :type engine: FEM_2D.Engine, FEM_3D.Engine
        :param physical_group_id: Physical group ID of the surface (GMSH)
        :type physical_group_id: int
        :param M_full: The mass matrix.
        :type M_full: CSC Matrix
        :param C_full: The "viscous damping" matrix.
        :type C_full: CSC Matrix
        :param K_full: The stiffness matrix.
        :type K_full: CSC Matrix
        :param f: Function of the forcing function, defaults to None
        :type f: function(time_index): Array, optional
        :param g: Function of the inhomogeneous dirichlet function, defaults to None
        :type g: function(time_index): Array, optional
        :return: M,C,K,f_ret,G_Boundary,dof_interior_idx,dof_boundary_idx
        :rtype: (CSC, CSC, CSC, function(time_index): Array, function(time_index): Array, Array, Array)
        """
        
        #Get DOFs which are not in the Dirichlet B.
        dof_interior_idx = np.argwhere(engine.mesh.vertice_group!=physical_group_id)[:,0]

        #Remove DOFs from boundary
        M = M_full[dof_interior_idx,:]
        M = M[:,dof_interior_idx]
        C = C_full[dof_interior_idx,:]
        C = C[:,dof_interior_idx]
        K = K_full[dof_interior_idx,:]
        K = K[:,dof_interior_idx]

        #Forcing function
        f = engine.F_Matrix(f)
        F = lambda time_index: f(time_index)[dof_interior_idx]

        #Get only the DOFs from the boundary and load the value of the function
        dof_boundary_idx = np.argwhere(engine.mesh.vertice_group==physical_group_id)[:,0]
        G_Boundary = lambda time_index: g(time_index)[dof_boundary_idx]

        #Extra term that handles the Inhomogeneous Dirichlet B.C.
        A_Boundary = (K_full[dof_interior_idx,:]+ M_full[dof_interior_idx,:])[:,dof_boundary_idx]
        G = lambda time_index: A_Boundary.dot(G_Boundary(time_index))

        #Add the extra term to the RHS
        f_ret = lambda time_index: F(time_index) - G(time_index)

        return M,C,K,f_ret,G_Boundary,dof_interior_idx,dof_boundary_idx

    @staticmethod
    def Impedance_Damping_Matrix(m3d, physical_group_id):
        """Calculates the damping matrix for surface impedance boundary condition, given the physical group(GMSH) of the surface. Only 3D (Lagrange P1)

        :param m3d: Mesh instance in 3D domain.
        :type m3d: FEM_3D.Mesh
        :param physical_group_id: Physical group id(GMSH) of the impedance surface.
        :type physical_group_id: int
        :return: The mesh instance of the surface in 2D, the nodetags of the 3D domain for each 2D dof. 
        :rtype: FEM_2D.Mesh, Array
        """
        triangles,vertices = Boundary.getPhysicalGroupTriangles(m3d,physical_group_id)

        unique, unique_indexes, unique_inverse = np.unique(triangles,return_index=True,return_inverse=True)
        nodeTags = unique
        vertices = vertices[unique_indexes]
        triangles = unique_inverse.reshape(-1,3)

        m2d = fem2d.Mesh.MeshByTriangles('ImpedanceSurface',nodeTags,triangles,vertices)
        engine2d = fem2d.Engine(m2d,1,1,calcForMass3D=True)
        M_2d = engine2d.M_Matrix().tocoo()

        gmsh.model.remove()
        return M_2d,nodeTags
        
    @staticmethod
    def getPhysicalGroupTriangles(m, physical_group_id):
        """Return the triangles that belongs to the surface of given physical group(GMSH).

        :param m: The mesh instance.
        :type m: FEM_2D.Mesh, FEM_3D.Mesh
        :param physical_group_id: Physical group id(GMSH)
        :type physical_group_id: int
        :return: Array with the vertices(indexes), Array of vertice coordinates.
        :rtype: Array, Array
        """
        nodes = m.nodeTags
        nodes_coords = m.vertices

        surface_tags = m.model.getEntitiesForPhysicalGroup(2,physical_group_id)
        triangles=[]
        vertices=[]
        for st in surface_tags:
            elements = m.model.mesh.getElements(2,st)
            _triangles = elements[2][0]
            _vertices = nodes_coords[np.where(_triangles[:,None]==nodes)[1]]
            vertices.extend(_vertices)
            triangles.extend(_triangles)

        return np.array(triangles),np.array(vertices)





import numpy as np
from . import TimeEngine,Boundary,Sources
from . import FEM_3D as fem3d
import gmsh
import scipy.sparse as sparse
import scipy.signal as signal
import scipy.fft as fft
import scipy.io.wavfile as wavfile
import time
import pickle as pkl

#TODO: Userfriendly errors and instructions
class Simulation():
    """Class that handles the calculation of the room response (time and frequency) for a 
    given mesh with defined impedances at a surface.
    """

    def __init__(self,simulationsTime,nameSource,nameReceivers):
        """Initialize the simulation configurations.

        :param simulationsTime: Dictionary of bands with the simulation times.
        :type simulationsTime: Dict(String,float)
        :param nameSource: Physical group name(GMSH) of the source node.  
        :type nameSource: string
        :param nameReceivers: Physical group name(GMSH) of the receivers nodes. 
        :type nameReceivers: Array(string)
        """
        self.nameSource = nameSource
        self.nameReceivers = nameReceivers
        self.simulationsTime = simulationsTime #TODO: Check sizes before simulation

        #Default parameters
        self.c = 340.
        self.cfl = .8
        self.elementsPerWavelength = 6
        self.element_order = 1
        self.max_h = 1e10

        #Initializations
        self.impedancesByBand = {}
        self.GroupTags = {}
        self.outputPath = None
        self.refine_BC = False

        #Default bands
        self.bands_freq = {
        '50'   : [ 44.194 , 55.681 ],
        '63'   : [ 55.681 , 70.154 ],
        '80'   : [ 70.154 , 88.388 ],
        '100'  : [ 88.388 , 111.362 ],
        '125'  : [ 111.362 , 140.308 ],
        '160'  : [ 140.308 , 176.777 ],
        '200'  : [ 176.777 , 222.725 ],
        '250'  : [ 222.725 , 280.616 ],
        '315'  : [ 280.616 , 353.553 ],
        '400'  : [ 353.553 , 445.449 ],
        '500'  : [ 445.449 , 561.231 ],
        '630'  : [ 561.231 , 707.107 ],
        '800'  : [ 707.107 , 890.899 ]} 

    def setFrequencyBands(self,bands):
        """Change the dictionary of bands and its cut-off frequencies.

        :param bands: Dictionary of the band edges
        :type bands: Dict(string,Array)
        """
        self.bands_freq = bands

    def setParameters(self,c=340.,elementsPerWavelength=6,cfl=.8):
        """Change the default parameters for sound speed, elements per wavelength and the CFL condition number.

        :param c: Sound speed, defaults to 340.
        :type c: float, optional
        :param elementsPerWavelength: Number of elements per wavelength, defaults to 6
        :type elementsPerWavelength: int, optional
        :param cfl: CFL condition number, defaults to .8
        :type cfl: float, optional
        """
        self.c = c
        self.elementsPerWavelength = elementsPerWavelength
        self.cfl = cfl

    def setMaxStepSize(self, h):
        """Define a max-size for the elements size.

        :param h: Elements size
        :type h: float
        """
        self.max_h = h

    def setBCRefinement(self, max_h):
        """Set specific mesh-size for the meshes at boundaries.

        :param max_h: Maximum size for the elements in the boundaries.
        :type max_h: float
        """
        self.refine_BC = True
        self.BC_max_h = max_h

    def setAbsorption(self, physicalGroupName, absorptionCoefsByBand):
        """For a given surface name, set the absorption coefficients for each band.

        :param physicalGroupName: Physical group name(GMSH) of the impedance surface
        :type physicalGroupName: string
        :param absorptionCoefsByBand: Dictionary with the absorption coef. for each band
        :type absorptionCoefsByBand: Dict(string,float)
        """
        physicalGroupName = physicalGroupName.upper()
        self.impedancesByBand[physicalGroupName] = {}
        for band in absorptionCoefsByBand:
            ref_coef=np.sqrt(1-absorptionCoefsByBand[band])
            impedancesByBand = (1+ref_coef)/(1-ref_coef)
            self.impedancesByBand[physicalGroupName][band] = impedancesByBand

    def setImpedance(self, physicalGroupName, impedancesByBand):
        """For a given surface name, set the impedance for each band.

        :param physicalGroupName: Physical group name(GMSH) of the impedance surface
        :type physicalGroupName: string
        :param impedancesByBand: Dictionary with the complex impedance for each band
        :type impedancesByBand: Dict(string,float)
        """
        physicalGroupName = physicalGroupName.upper()
        self.impedancesByBand[physicalGroupName] = impedancesByBand

    def readGeometryDXF(self,file):
        """Read a geometry described in the DXF format.

        :param file: [description]
        :type file: [type]
        """
        m = fem3d.Mesh(file)
        m.readDXF(file)
        self.setMesh(m)

    def readGeometryGEO(self,file):
        """Read a geometry described in GEO format.

        :param file: [description]
        :type file: [type]
        """
        m = fem3d.Mesh(file)
        m.readGeo(file)
        self.setMesh(m)

    def setMesh(self, mesh: fem3d.Mesh):
        """Set the mesh instance and load the surfaces.

        :param mesh: [description]
        :type mesh: FEM_3D.Mesh
        """
        self.mesh = mesh
        for dimTag in self.mesh.model.getPhysicalGroups():
            self.GroupTags[self.mesh.model.getPhysicalName(dimTag[0],dimTag[1])] = dimTag

    def setOutputPath(self, path):
        """Set the output to the WAV files from the source and receivers.

        :param path: [description]
        :type path: [type]
        """
        self.outputPath = path

    def run(self): #TODO: Write status?
        """Run the simulation for each band and stores its results.

        :return: [description]
        :rtype: [type]
        """
        source_signals = {}
        time_arrays = {}
        receiver_signals = {}
        t1 = time.time()
        self.saveParameters()
        groups_tags = self.GroupTags
        
        for band in self.bands_freq:
            t1_band = time.time()

            band_range = np.array(self.bands_freq[band])
            print('Starting band ' + band)
            self.mesh.model.mesh.clear()

            #Set the parameters specific for the band frequencies
            wavelengths=self.c/band_range
            h=np.min(wavelengths)/self.elementsPerWavelength
            h=np.min([h,self.max_h])
            dt=self.cfl*h/self.c #CFL Condition
            tf = self.simulationsTime[band]

            if tf==0: continue

            self.mesh.fac.synchronize()
            pTags = self.mesh.model.getEntities(0)
            self.mesh.model.occ.setMeshSize(pTags,h)
            self.mesh.model.mesh.setSize(pTags,h)

            #Impedance Surfaces Refinement
            if self.refine_BC:
                h_bc = np.min([h,self.BC_max_h])
                for material in self.impedancesByBand:
                    material_tag = groups_tags[material]
                    materialSurfaceTag = self.mesh.model.getEntitiesForPhysicalGroup(material_tag[0],material_tag[1])
                    pSurfaceTag = self.mesh.model.getBoundary([(2,i) for i in materialSurfaceTag],recursive=True)
                    self.mesh.model.occ.setMeshSize(pSurfaceTag,h_bc)
                    self.mesh.model.mesh.setSize(pSurfaceTag,h_bc)

            self.mesh.fac.synchronize()
            self.mesh.generate()

            #Assemble the FEM matrices
            engine = fem3d.Engine(self.mesh,self.element_order,self.element_order)
            ndofs = len(engine.dof)
            print('h: ',str(h), ' -- dofs:', ndofs)
            M = engine.M_Matrix()/(self.c**2)
            C = sparse.csc_matrix(M.shape)
            K = engine.K_Matrix()

            #Impedance B.C
            material_impedances = self.impedancesByBand
            for material in self.impedancesByBand:
                Z = material_impedances[material][band]
                material_tag = groups_tags[material][1]

                M_2d,nodeTags_imp = Boundary.Impedance_Damping_Matrix(self.mesh,material_tag)
                nodeTags_idx = np.where(nodeTags_imp[:,None]==self.mesh.nodeTags)[1]
                idx_row = nodeTags_idx[M_2d.row]
                idx_col = nodeTags_idx[M_2d.col]
                C_material = sparse.coo_matrix((M_2d.data, (idx_row, idx_col)), M.shape).tocsc()
                C_material = C_material/(self.c*Z)

                C = C + C_material

            #DOF index of the source:
            sourceDimTag = groups_tags[self.nameSource]
            sourceNodeTag = self.mesh.model.getEntitiesForPhysicalGroup(sourceDimTag[0],sourceDimTag[1])
            idx_xS = np.where(self.mesh.nodeTags==sourceNodeTag)[0]

            #DOF indexes of the receivers
            idx_xR = []
            for name in self.nameReceivers:
                receiverDimTag = groups_tags[name]
                receiverNodeTag = self.mesh.model.getEntitiesForPhysicalGroup(receiverDimTag[0],receiverDimTag[1])
                idx_xR.extend(np.where(self.mesh.nodeTags==receiverNodeTag)[0])

            #Setup the time scheme
            tspan = np.arange(0,tf,dt)
            # tengine = TimeEngine.Newmark(M,C,K,dt,alpha=.25,delta=.5)
            tengine = TimeEngine.Newmark_iterative(M,C,K,dt,alpha=.25,delta=.5)

            time_monopole = Sources.monopole_by_band(tspan,band_range)

            select_dof = np.zeros(ndofs)
            select_dof[idx_xS] = 1
            F = lambda time_index: time_monopole[time_index] * select_dof

            s,s_main = tengine.solve(tspan,F,idx_xR,[],1)
            receiver_signals[band] = s_main.T
            time_arrays[band] = tspan
            source_signals[band] = - time_monopole/(4*np.pi) #TODO Describe the negative sign

            self.saveOutput(band, self.nameSource,source_signals[band],receiver_signals[band],16000)

            t2_band = time.time()
            print('Simulation time (' + band + '): ', str(t2_band-t1_band))
        t2=time.time()
        print('Total simulation time: ', str(t2-t1))
        self.time_arrays = time_arrays
        self.source_signals = source_signals
        self.receiver_signals = receiver_signals
        return time_arrays,self.source_signals,self.receiver_signals
    
    def saveParameters(self):
        """Store in a file parameters.txt the parameters used in the simulation.
        """
        if (self.outputPath!=None):
            with open(self.outputPath + 'parameters.txt', 'a') as the_file:
                the_file.write('Source:' + str(self.nameSource) + '\n')
                the_file.write('Receivers:' + str(self.nameReceivers) + '\n')
                the_file.write('Elements per Wavelength:' + str(self.elementsPerWavelength) + '\n')
                the_file.write('Bands:' + str(self.bands_freq) + '\n')
                the_file.write('Simulation times:' + str(self.simulationsTime) + '\n')
                the_file.write('Impedances:' + str(self.impedancesByBand) + '\n')

    def saveOutput(self, band, source_name, source_signal, receiver_signals, save_fs): 
        """Save the signals as WAV. Source and Receivers for each band.

        :param band: [description]
        :type band: [type]
        :param source_name: [description]
        :type source_name: [type]
        :param source_signal: [description]
        :type source_signal: [type]
        :param receiver_signals: [description]
        :type receiver_signals: [type]
        :param save_fs: [description]
        :type save_fs: [type]
        """
        if (self.outputPath!=None):
            source_signal = signal.resample(source_signal,int(self.simulationsTime[band]*save_fs)).astype(np.float32)
            normalize_by = np.max(source_signal)
            source_signal = source_signal/normalize_by
            wavfile.write(self.outputPath + str(band) + "_" + source_name + ".wav",save_fs,source_signal) 
            
            for i_rec in range(0,len(self.nameReceivers)):
                receiver_name = self.nameReceivers[i_rec]
                receiver_signal = receiver_signals[i_rec]
                receiver_signal = signal.resample(receiver_signal,int(self.simulationsTime[band]*save_fs)).astype(np.float32)
                receiver_signal = receiver_signal/normalize_by
                wavfile.write(self.outputPath + str(band) + "_" + receiver_name + ".wav",save_fs,receiver_signal) 

    def calcFrequencyResponse(self, receiverIndex = 0): #TODO EXPLAIN
        """ Return the frequency response given by the simulation for the specific receiver. """
        resampled_tspan = {}
        resampled_source_signals = {}
        resampled_receiver_signals = {}
        sources_fft = {}
        receivers_fft = {}

        #Resample every signal to the max sampling frequency of the bands
        max_band = str(np.max(list(map(int,self.source_signals.keys())))) #Band name/key

        #Get the max sampling frequency of the simulated signals
        max_fs = 1/(self.time_arrays[max_band][1]-self.time_arrays[max_band][0])

        #Resample all the signals from the source and the selected receiver 
        for band in self.source_signals:
            new_size = int(np.max(self.time_arrays[band])*max_fs)
            rsource, rtspan = signal.resample(self.source_signals[band],new_size,self.time_arrays[band])
            rreceiver = signal.resample(self.receiver_signals[band][receiverIndex],new_size)

            resampled_tspan[band] = rtspan
            resampled_source_signals[band] = rsource
            resampled_receiver_signals[band] = rreceiver
            sources_fft[band] = fft.rfft(rsource)
            receivers_fft[band] = fft.rfft(rreceiver)
        
        #Include the response for each band, only for its frequencies
        frequencies = []
        response = []
        for band in self.source_signals:
            _band_freqs = fft.rfftfreq(len(resampled_tspan[band]),1/max_fs)
            idx_from = np.argmin(np.abs(_band_freqs-self.bands_freq[band][0]))
            idx_to = np.argmin(np.abs(_band_freqs-self.bands_freq[band][1]))
            band_response = receivers_fft[band]/sources_fft[band]
            frequencies.extend(_band_freqs[idx_from:idx_to])
            response.extend(band_response[idx_from:idx_to])

        frequencies = np.array(frequencies)
        response = np.array(response)

        return frequencies, response

    def getImpulseResponse(self): #TODO Impulse response from any signal
        source_sum = np.zeros(max_size)
        rec_sum = np.zeros(max_size)
        bands_int = [50,63,80,100,125,160,200,250,315,400,500,630,800]
        for bidx in range(0,len(bands_int)):
            # if bidx==len(bands_int)-1: continue
            if bidx==0:
                bkey_low = str(bands_int[bidx])
                cutoff = np.min(self.bands_freq[bkey_low])
                ssig = self.source_signals[bkey_low]
                rsig = self.receiver_signals[bkey_low][0]
            else:
                bkey_low = str(bands_int[bidx-1])
                bkey_high = str(bands_int[bidx])
                cutoff = np.max(self.bands_freq[bkey_low])
                ssig = self.source_signals[bkey_high]
                rsig = self.receiver_signals[bkey_high][0]

            ssig_to_filt = np.zeros(max_size)
            ssig_to_filt[:len(ssig)] = ssig
            rsig_to_filt = np.zeros(max_size)
            rsig_to_filt[:len(rsig)] = rsig

            filter_low = signal.firwin(1001,cutoff,pass_zero='lowpass',fs=fs)
            filter_high = signal.firwin(1001,cutoff,pass_zero='highpass',fs=fs)
            source_sum = signal.filtfilt(filter_low,[1],source_sum) + signal.filtfilt(filter_high,[1],ssig_to_filt)
            rec_sum = signal.filtfilt(filter_low,[1],rec_sum) + signal.filtfilt(filter_high,[1],rsig_to_filt)


import numpy as np
import scipy.sparse as sparse
import scipy.sparse.linalg as sla
import sys
from . import Tools

class Newmark_iterative:
    # M ddu + C du + K u = f
    # a0 = 1/(alpha*dt**2)
    # a1 = delta/(alpha*dt)
    # a2 = 1/(alpha*dt)
    # a3 = 1/(2*alpha) - 1
    # a4 = delta/alpha - 1
    # a5 = dt/2 * (delta/(alpha-2))
    # a6 = dt*(1-delta)
    # a7 = delta*dt
    # Keff = K + a0*M + a1*C
    # R = f(t+dt) + M(a0*u + a2*du + a3*ddu) + C(a1*u + a4*du + a5*ddu)

    # Systeme : Keff u(t+dt) = R(t+dt)
    # delta > 0.5
    # alpha > 0.25(0.5 + delta)^2
    # Triangularize Keff
    def __init__(self,M,C,K,dt,alpha,delta):
        self.alpha = alpha
        self.delta = delta
        self.dt = dt
        self.a0 = 1/(alpha*(dt**2))
        self.a1 = delta/(alpha*dt)
        self.a2 = 1/(alpha*dt)
        self.a3 = 1/(2*alpha) - 1
        self.a4 = delta/alpha - 1
        self.a5 = (dt/2) * ((delta/alpha)-2)
        self.a6 = dt*(1-delta)
        self.a7 = delta*dt
        self.tspan = []
        self.u = []

        self.M = M
        self.C = C
        if 'umfpack' in sys.modules:
            sla.use_solver(useUmfpack=True)
        self.Keff = K + M.multiply(self.a0) + C.multiply(self.a1)
        self.Keff.indices = self.Keff.indices.astype(np.int)
        self.Keff.indptr = self.Keff.indptr.astype(np.int)

        #Initial conditions
        self.u0 = np.zeros(M.shape[0])
        self.du0 = np.zeros(M.shape[0])
        self.ddu0 = np.zeros(M.shape[0])

    def setInitialConditions(self,u0,du0,ddu0):
        self.u0 = u0
        self.du0 = du0
        self.ddu0 = ddu0

    def preconditionate(self,Keff):
        ilu = sla.spilu(Keff)
        Mx = lambda x: ilu.solve(x)
        K_prec = sla.LinearOperator(Keff.shape, Mx)
        return K_prec
        
    def solve(self,tspan,F,main_dofs,saved_dofs,save_time_step): #TODO: Check if the F_Matrix is being passed
        self.tspan=tspan

        Keff_solve = 0
        self.Keff_prec = self.preconditionate(self.Keff)

        ret_u = [self.u0[saved_dofs]]
        ret_main_dofs=[self.u0[main_dofs]]
        u = self.u0
        du = self.du0
        ddu = self.ddu0
        for i in range(1,len(tspan)):
            f_new = -F(i)

            u,du,ddu = self.iterate(Keff_solve,f_new,u,du,ddu)
            ret_main_dofs.append(u[main_dofs])

            if i % save_time_step < 1e-20:
                ret_u.append(u[saved_dofs])

            Tools.Other.printInline('Time Step: ' + str(i+1) + '/' + str(len(tspan)))

        ret_u = np.array(np.real(ret_u))
        ret_main_dofs = np.array(np.real(ret_main_dofs))
        Tools.Other.printInline('\n\r')
        self.u = ret_u
        return ret_u,ret_main_dofs

    def iterate(self,Keff_solve,f_new,u,du,ddu):
        R_new = f_new + self.M.dot(self.a0*u + self.a2*du + self.a3*ddu) + self.C.dot(self.a1*u + self.a4*du + self.a5*ddu)

        u_new = sla.gcrotmk(self.Keff,R_new,M=self.Keff_prec,x0=u,atol=1e-5)[0]

        ddu_new = self.a0*(u_new - u) - self.a2*du - self.a3*ddu
        du_new = du + self.a6*ddu + self.a7*ddu_new
        return u_new,du_new,ddu_new

class Newmark:
    # M ddu + C du + K u = f
    # a0 = 1/(alpha*dt**2)
    # a1 = delta/(alpha*dt)
    # a2 = 1/(alpha*dt)
    # a3 = 1/(2*alpha) - 1
    # a4 = delta/alpha - 1
    # a5 = dt/2 * (delta/(alpha-2))
    # a6 = dt*(1-delta)
    # a7 = delta*dt
    # Keff = K + a0*M + a1*C
    # R = f(t+dt) + M(a0*u + a2*du + a3*ddu) + C(a1*u + a4*du + a5*ddu)

    # Systeme : Keff u(t+dt) = R(t+dt)
    # delta > 0.5
    # alpha > 0.25(0.5 + delta)^2
    # Triangularize Keff
    def __init__(self,M,C,K,dt,alpha,delta):
        self.alpha = alpha
        self.delta = delta
        self.dt = dt
        self.a0 = 1/(alpha*(dt**2))
        self.a1 = delta/(alpha*dt)
        self.a2 = 1/(alpha*dt)
        self.a3 = 1/(2*alpha) - 1
        self.a4 = delta/alpha - 1
        self.a5 = (dt/2) * ((delta/alpha)-2)
        self.a6 = dt*(1-delta)
        self.a7 = delta*dt
        self.tspan = []
        self.u = []

        self.M = M
        self.C = C
        if 'scikits.umfpack' in sys.modules:
            sla.use_solver(useUmfpack=True)
        self.Keff = K + M.multiply(self.a0) + C.multiply(self.a1)
        self.Keff.indices = self.Keff.indices.astype(np.int64)
        self.Keff.indptr = self.Keff.indptr.astype(np.int64)

        #Initial conditions
        self.u0 = np.zeros(M.shape[0])
        self.du0 = np.zeros(M.shape[0])
        self.ddu0 = np.zeros(M.shape[0])

    def setInitialConditions(self,u0,du0,ddu0):
        self.u0 = u0
        self.du0 = du0
        self.ddu0 = ddu0
        
    def solve(self,tspan,f,main_dofs,saved_dofs,save_time_step):
        self.tspan=tspan

        Keff_solve = sla.factorized(self.Keff)

        ret_u = [self.u0[saved_dofs]]
        ret_main_dofs=[self.u0[main_dofs]]
        u = self.u0
        du = self.du0
        ddu = self.ddu0
        for i in range(1,len(tspan)):
            f_new = -f(i)

            u,du,ddu = self.iterate(Keff_solve,f_new,u,du,ddu)
            ret_main_dofs.append(u[main_dofs])

            if i % save_time_step < 1e-20:
                ret_u.append(u[saved_dofs])

            Tools.Other.printInline('Time Step: ' + str(i+1) + '/' + str(len(tspan)))

        ret_u = np.array(ret_u)
        ret_main_dofs = np.array(ret_main_dofs)
        Tools.Other.printInline('\n\r')
        self.u = ret_u
        return ret_u,ret_main_dofs

    def iterate(self,Keff_solve,f_new,u,du,ddu):
        R_new = f_new + self.M.dot(self.a0*u + self.a2*du + self.a3*ddu) + self.C.dot(self.a1*u + self.a4*du + self.a5*ddu)

        u_new = Keff_solve(R_new)

        ddu_new = self.a0*(u_new - u) - self.a2*du - self.a3*ddu
        du_new = du + self.a6*ddu + self.a7*ddu_new
        return u_new,du_new,ddu_new

class LeapFrog():
    #(M + dt*C + (dt**2)*K) u_new = (2M + dt*C) u1 - M u0 - (dt**2)f[n+1]
    def __init__(self,M,C,K,dt):
        self.dt = dt
        self.tspan = []
        self.u = []

        self.M = M
        self.C = C
        self.K = K
        if 'scikits.umfpack' in sys.modules:
            sla.use_solver(useUmfpack=True)
        self.sys = M + dt*C + (dt**2)*K
        self.sys.indices = self.sys.indices.astype(np.int64)
        self.sys.indptr = self.sys.indptr.astype(np.int64)

        self.a0 = self.M.multiply(2) + self.C.multiply(self.dt)
        self.a1 = self.dt**2

    def solve(self,tspan,f,u0,du0,ddu0,main_dofs,saved_dofs,save_time_step):
        self.tspan=tspan
        sys_solve = sla.factorized(self.sys)
        ret_u = [u0[saved_dofs]]
        ret_main_dofs=[u0[main_dofs]]
        u0 = u0
        u1 = u0
        for i in range(1,len(tspan)):
            f_new = -f(i)
            u_new = self.iterate(sys_solve,f_new,u0,u1)
            ret_main_dofs.append(u_new[main_dofs])
            if i % save_time_step < 1e-20:
                ret_u.append(u_new[saved_dofs])
            u0=u1
            u1=u_new

        ret_u = np.array(ret_u)
        ret_main_dofs = np.array(ret_main_dofs)
        self.u = ret_u
        return ret_u,ret_main_dofs

    def iterate(self,sys_solve,f_new,u0,u1):
        R_new = self.a0*u1 - self.M.dot(u0) - self.a1*f_new
        u_new = sys_solve(R_new)
        return u_new


import numpy as np
import numpy.linalg as la

def calcL2(u,omega,result,exact):
    """For a given result and the exact solution, returns the L2-norm of the error.

    :param u: The support matrix u, with the values of the interpolating functions at the integration points.
    :type u: CSC Matrix
    :param omega: Weights of the quadrature formula at the integration points
    :type omega: Array
    :param result: Result given from the solver. Value at each degree of freedom.
    :type result: Array
    :param exact: Exact solution at each degree of freedom
    :type exact: Array
    :return: The L2-norm of the error.
    :rtype: float
    """
    integrand = (exact-result)**2
    return np.sqrt(u.dot(integrand).dot(omega))

def calcH1(u,dxu,dyu,omega,result,exact,grad_exact):
    """For a given result and the exact solution, returns the H1-norm of the error.

    :param u: The support matrix u, with the values of the interpolating functions at the integration points.
    :type u: CSC Matrix
    :param dxu: The support matrix dxu, with the x derivative values
    :type dxu: CSC Matrix
    :param dyu: The support matrix dyu, with the x derivative values
    :type dyu: CSC Matrix
    :param omega: Weights of the quadrature formula at the integration points
    :type omega: Array
    :param result: Result given from the solver. Value at each degree of freedom.
    :type result: Array
    :param exact: Exact solution at each degree of freedom
    :type exact: Array
    :param grad_exact: Gradient of the xact solution at each degree of freedom
    :type grad_exact: Array
    :return: The H1-norm of the error.
    :rtype: float
    """
    integral1 = u.dot((exact-result)**2).dot(omega)
    dif_grad = np.array(u.dot(grad_exact)) - np.array([dxu.dot(result),dyu.dot(result)]).transpose()
    integral2 = (la.norm(dif_grad,axis=1)*la.norm(dif_grad,axis=1)).dot(omega)
    return np.sqrt(integral1 + integral2)


from .Boundary import Boundary
from .Tools import Visualization,Sources,SaveLoad,Other
from .Simulation import Simulation
from .Solver import Solver
from .TimeEngine import Newmark,LeapFrog
try:
    import scikits.umfpack#Recommended: scikit-umfpack  (0.3.1 +)
except:
    pass




import sys
import numpy as np
import scipy.sparse as sparse
import scipy.signal as signal
import scipy.fft as fft
import pickle as pk
import gmsh
import numpy.linalg as la
from . import TimeEngine,Boundary
from . import FEM_3D as fem3d
from . import FEM_2D as fem2d

class Sources():

    @staticmethod
    def monopole_by_band(t_,band,sigma=None):
        """ Return a forcing function representing a monopole source with max amplitude 1 at 1 meter. 
        The signal is a gaussian with the given cutoff frequencies (3dB). """
        freq = (np.max(band)+np.min(band))/2 #Center frequency
        omega = 2*np.pi*freq

        if sigma==None:
            fc = (np.max(band)-np.min(band))
            sigma=1/(2*np.pi*fc/4) #TODO: what is the best bandwidth?

        t=t_-5*sigma #Shifts the gaussian so it can start at zero.

        _signal = np.sin(omega*t) * np.exp(-t**2/(2*sigma**2))
        _signal = _signal/np.max(np.abs(_signal))
        return 4 * np.pi * _signal

class Visualization():
    @staticmethod
    def addView(mesh,viewName,timeData,nodeTags):
        """timeData must be (Timesteps)x(#dof)"""
        viewTag = mesh.pos.add(viewName)
        for i_timeData in range(0,len(timeData)):
            mesh.pos.addModelData(viewTag,i_timeData,mesh.name,'NodeData',nodeTags,timeData[i_timeData].reshape(-1,1),numComponents=1)

    @staticmethod
    def showTime(mesh,viewNames,data,view_dofs,fltkrun=True):
        """data must be (Views)x(Timesteps)x(#dof)"""
        for i_view in range(0,len(data)):
            nodeTags = mesh.nodeTags[view_dofs[i_view]]
            timeData = data[i_view]
            Visualization.addView(mesh,viewNames[i_view],timeData,nodeTags)

        if (fltkrun):
            mesh.FLTKRun()

class SaveLoad():
    def save(self,name,saved_dofs,dof,sln,sln_main,tspan):
        """Save the mesh and the main infos about the numerical result."""
        self.dof = dof
        self.sln = sln
        self.sln_main = sln_main
        self.tspan = tspan
        self.saved_dofs = saved_dofs
        gmsh.option.setNumber('Mesh.SaveAll',1)
        gmsh.write('data/' + name + '.msh')
        with open('data/' + name + '.pickle', 'wb') as f:
            pk.dump(self, f, pk.HIGHEST_PROTOCOL)

    def load(self,name,m):
        """Load the mesh and the main infos about a numerical result."""
        gmsh.open('data/' + name + '.msh')
        m.readMsh('data/' + name + '.msh')
        with open('data/' + name + '.pickle', 'rb') as f:
            T = pk.load(f)
            self.dof = T.dof
            self.sln = T.sln
            self.sln_main = T.sln_main
            self.tspan = T.tspan
            self.saved_dofs = T.saved_dofs

class Other():
    @staticmethod
    def nearest_dof(dofs,x): #TODO: Not here
        """Return the index of the nearest 3D degree of freedom"""
        idx = np.argmin(la.norm(dofs - x,axis=1))
        return idx

    @staticmethod
    def printInline(text):
        sys.stdout.write(text + '                                            \r')

    @staticmethod
    def wiener_deconvolution(output, input, SNR=1):
        input = np.hstack((input, np.zeros(len(output) - len(input)))) # zero pad the kernel to same length
        H = fft.fft(input)
        deconvolved = np.real(fft.ifft(fft.fft(output)*np.conj(H)/(H*np.conj(H) + SNR**2)))
        return deconvolved

    @staticmethod
    def resample(signal,from_fs,to_fs,from_duration,to_duration):#TODO Resample
        return 

# class RoomResponse():
#     """ Class that handles the calculation of the impulse response. """

#     def __init__(self,M,C,K,c,dt,tf,f_band,idx_xS,idx_xR):
#         self.M = M
#         self.C = C
#         self.K = K
#         self.c = c
#         self.dt = dt
#         self.tf = tf
#         self.f_band = f_band
#         self.idx_xS = idx_xS  #TODO: uses the interpolation to get the value
#         self.idx_xR = idx_xR
#         self.ndofs = M.shape[0]
        
#         self.tspan = np.arange(0,self.tf,self.dt)
#         self.tengine = TimeEngine.Newmark_iterative(self.M,self.C,self.K,self.dt,alpha=.25,delta=.5)

#         self.time_monopole = Sources.monopole_by_band(self.tspan,self.f_band)

#         select_dof = np.zeros(self.ndofs)
#         select_dof[idx_xS] = 1
#         self.forcing_function = lambda time_index: self.time_monopole[time_index] * select_dof

#     def run(self):
#         """ Calculate the room response and return: Discretized frequencies, input spectrum, output spectrum, frequency response, output time response. """
#         s,s_main = self.tengine.solve(self.tspan,self.forcing_function,self.idx_xR,[],1)

#         all_frequencies=fft.rfftfreq(len(self.tspan),self.dt)
#         valid_band_idx = np.arange((np.abs(all_frequencies-self.f_band[0])).argmin(),(np.abs(all_frequencies-self.f_band[1])).argmin()+1)
#         frequencies = all_frequencies[valid_band_idx]
#         pressure1m = np.array([self.forcing_function(i)[self.idx_xS] for i in range(0,len(self.tspan))])/(4*np.pi) # Forcing divided by 4pi so the reference pressure be at distance 1
#         f_pressure1m = fft.rfft(pressure1m)[valid_band_idx]
#         f_result = []
#         f_ratio = []
#         for idx in range(0,len(self.idx_xR)):
#             f_result.append(fft.rfft(s_main[:,idx])[valid_band_idx])
#             f_ratio.append(fft.rfft(s_main[:,idx])[valid_band_idx]/f_pressure1m) #

#         return frequencies,f_pressure1m,f_result,f_ratio,s_main


import numpy as np
import scipy.sparse as sparse
import sys


class Engine:
    def __init__(self,mesh,order,formule,calcForMass3D=False):
        """Initialize the engine and assembles the support matrices 

        :param mesh: Instance of the mesh.
        :type mesh: FEM_2D.Mesh
        :param order: Order of the elements. [1 or 2]
        :type order: int
        :param formule: Formule for the numerical integration. Recommended >= (order)
        :type formule: int
        :param calcForMass3D: True if the mesh is in 3D space. Applies only for assembling the mass matrix. defaults to False
        :type calcForMass3D: bool, optional
        """
        self.order = order
        self.formule = formule
        self.mesh = mesh
        self.calcForMass3D = calcForMass3D
        self.dof,self.numDof,self.u,self.dxu,self.dyu,self.omega = self.initializeMatrices('Lagrange')
        
    def IntegrationPointsHat(self,formule: int):
        """Return the integration points of the reference triangle

        :param formule: Formule for the integration. [1 to 6]
        :type formule: int
        :return: (x,y,omega) ; x and y being the coordinates at the reference domain. Omega the weight of integration.
        :rtype: Tuple of arrays
        """
        xhat = []
        yhat = []
        omega = []
        if (formule==1):
            xhat = [1/3]
            yhat = [1/3]
            omega = [1/2]
        elif (formule==2):
            xhat = [1/2,1/2,0]
            yhat = [1/2,0,1/2]
            omega = [1/6,1/6,1/6]
        elif (formule==3):
            xhat = [1/6,1/6,2/3]
            yhat = [1/6,2/3,1/6]
            omega = [1/6,1/6,1/6]
        elif (formule==4):
            xhat = [1/3,1/5,3/5,1/5]
            yhat = [1/3,1/5,1/5,3/5]
            omega = [-9/32,25/96,25/96,25/96]
        elif (formule==5):
            xhat = [1/2,1/2,0,1/6,1/6,2/3]
            yhat = [1/2,0,1/2,1/6,2/3,1/6]
            omega = [1/60,1/60,1/60,3/20,3/20,3/20]
        elif (formule==6):
            a = (6 + np.sqrt(15))/21
            b = (6 - np.sqrt(15))/21
            A = (155 + np.sqrt(15))/2400
            B = (155 - np.sqrt(15))/2400
            xhat = [1/3,a,1-2*a,a,b,1-2*b,b]
            yhat = [1/3,a,a,1-2*a,b,b,1-2*b]
            omega = [9/80,A,A,A,B,B,B]
        return (np.array(xhat),np.array(yhat),np.array(omega))

    def initializeMatrices(self, typeEF):
        """Generate the matrices with the values of the base functions, the dx, the dy and the weights for integration.

        :param typeEF: Type of inteporlation. Only 'Lagrange'.
        :type typeEF: string
        :return: (dof,numDof,u,dxu,dyu,omega);
        dof is the position of the degrees of freedom
        numDof is the dofs that composes each element
        u is the value of the interpolation function for each integration point
        dxu is the derivative in x
        dyu is the derivative in y
        omega is the quadrature weight for each integration point
        :rtype: (Array n x 3,Array e x 3,Matrix i x e,Matrix i x e,Array i); 
        n being the number of degrees of freedom; 
        e being the number of elements;
        i being the number of integration points
        """
        mesh,order,formule = self.mesh, self.order, self.formule
        xhat, yhat, omegaloc = self.IntegrationPointsHat(formule)

        if (typeEF=='Lagrange'):
            phi, dxphi, dyphi = self.LagrangeFunctions(xhat, yhat, order)

            if (order==0):
                for triangle in mesh.triangles:
                    coord_v1 = mesh.vertices[triangle[0]]
                    coord_v2 = mesh.vertices[triangle[1]]
                    coord_v3 = mesh.vertices[triangle[2]]

                    dof.append([(coord_v1[0]+coord_v2[0]+coord_v3[0])/3,
                                (coord_v1[1]+coord_v2[1]+coord_v3[1])/3])
                    numDof.append([len(dof)-1])
                numDof = np.array(numDof)
                dof = np.array(dof)
            elif (order==1 or order==2): #TODO: GMSH generate the nodes
                dof = mesh.vertices
                numDof = np.array(mesh.triangles,dtype=np.int32)

            if (not self.calcForMass3D):
                mesh.calcJacobians()
                areas_ratio = mesh.determinants
                J_inv = mesh.inverses
                dxhat_dx = J_inv[:,0,0]
                dxhat_dy = J_inv[:,0,1]
                dyhat_dx = J_inv[:,1,0]
                dyhat_dy = J_inv[:,1,1]
            else: #If the coordinates are in 3D uses another formula
                areas_ratio = mesh.calcAreasRatio3D()
                
            N_dof = len(dof)
            N_int = len(mesh.triangles)*len(xhat)
            omega = np.zeros(N_int)

            #The sparse matrices will have for each line a integration point and each
            #column a dof. The number of integration points is the number of elements
            #times the number of points per element. So, for each point of integration
            #we have a certain amount of base functions to evaluate, the amount of 
            #dof per element.
            #Structure for sparse Matrix
            _i = np.zeros(N_int*len(phi[0]))
            _j = np.zeros(N_int*len(phi[0]))
            _v_u = np.zeros(N_int*len(phi[0]))
            _v_dxu = np.zeros(N_int*len(phi[0]))
            _v_dyu = np.zeros(N_int*len(phi[0]))

            for iloc in range(0,len(xhat)): #Points of Integration
                iglob = np.arange(iloc,N_int,len(xhat),dtype=np.int32)
                omega[iglob] = omegaloc[iloc] * areas_ratio
                for jloc in range(0,len(phi[0])): #Base functions
                    _idx = jloc*N_int + iloc*len(iglob) + np.arange(0,len(iglob)) #Global stacked index for the sparse matrix
                    jglob = numDof[:,jloc]
                    _i[_idx] = iglob
                    _j[_idx] = jglob
                    _v_u[_idx] = np.ones(len(iglob))*phi[iloc,jloc]

                    if not self.calcForMass3D: #The derivatives are not needed for the mass matrix
                        _v_dxu[_idx] = (dxphi[iloc,jloc]*dxhat_dx + dyphi[iloc,jloc]*dyhat_dx)
                        _v_dyu[_idx] = (dxphi[iloc,jloc]*dxhat_dy + dyphi[iloc,jloc]*dyhat_dy)
                    
            #Generate the matrices from the row/col indexes and values
            u = sparse.coo_matrix((_v_u, (_i, _j)),(N_int,N_dof))
            dxu = sparse.coo_matrix((_v_dxu, (_i, _j)),(N_int,N_dof))
            dyu = sparse.coo_matrix((_v_dyu, (_i, _j)),(N_int,N_dof))
            u = u.tocsc()
            dxu = dxu.tocsc()
            dyu = dyu.tocsc()

        return (np.array(dof),np.array(numDof),u,dxu,dyu,omega)

    def LagrangeFunctions(self, x, y, order):
        """Get the values and of the base functions and its derivatives at the triangle of reference.

        :param x: Array of x positions in the domain of reference
        :type x: array(float)
        :param y: Array of y positions in the domain of reference
        :type y: array(float)
        :param order: Order of the Lagrange function. [1 or 2]
        :type order: int
        :raises Exception: If the Lagrange function order are no implemented
        :return: (value,x derivative,y derivative)
        :rtype: (array(float),array(float),array(float))
        """
        phi = []
        dxphi = []
        dyphi = []
        if (order==0):
            for i in range(0,len(x)):
                phi.append([1.])
                dxphi.append([0.])
                dyphi.append([0.])
        elif (order==1): #Ordering (0,0), (1,0) et (0,1)
            for i in range(0,len(x)):
                phi.append([1-x[i]-y[i],x[i],y[i]])
                dxphi.append([-1.,1.,0.])
                dyphi.append([-1.,0.,1.])
        elif (order==2): #Ordering (0,0), (1,0), (0,1), (.5,0), (.5,.5) et (0,.5)
            for i in range(0,len(x)):
                phi.append([(1-x[i]-y[i])*(1-2*x[i]-2*y[i]),
                            x[i]*(2*x[i]-1),
                            y[i]*(2*y[i]-1),
                            4*x[i]*(1-x[i]-y[i]),
                            4*x[i]*y[i],
                            4*y[i]*(1-x[i]-y[i])])
                dxphi.append([4*y[i]+4*x[i]-3,
                                4*x[i]-1,            
                                0,
                                -4*(2*x[i] + y[i] - 1),          
                                4*y[i],
                                -4*y[i]])
                dyphi.append([4*y[i]+4*x[i]-3,    
                                0,            
                                4*y[i]-1,               
                                -4*x[i],            
                                4*x[i],                
                                -4*(x[i] + 2*y[i] - 1)])
        else: 
            raise Exception('Lagrande functions of order ' + str(order) + ' not implemented.')
        return np.array(phi), np.array(dxphi), np.array(dyphi)
    
    def M_Matrix(self):
        """ Generate the mass matrix from the support matrices.

        :return: The mass matrix.
        :rtype: CSC Matrix
        """
        u,omega  = self.u,self.omega
        M = u.transpose().dot(u.transpose().multiply(omega).transpose())
        M = M.tocsc()

        #Garantee int64 to UMFPACK
        
        if 'scikits.umfpack' in sys.modules:
            M.indices = M.indices.astype(np.int64)
            M.indptr = M.indptr.astype(np.int64)
        return M

    def K_Matrix(self):
        """ Generate the stiffness matrix from the support matrices.

        :return: The stiffness matrix.
        :rtype: CSC Matrix
        """
        dxu,dyu,omega  = self.dxu,self.dyu,self.omega
        G1 = dxu.transpose().dot(dxu.transpose().multiply(omega).transpose())
        G2 = dyu.transpose().dot(dyu.transpose().multiply(omega).transpose())
        K = G1+G2
        K = K.tocsc()

        #Garantee int64 to UMFPACK
        K.indices = K.indices.astype(np.int64)
        K.indptr = K.indptr.astype(np.int64)
        return K

    def F_Matrix(self, f):
        """Generate the forcing matrix.

        :param f: The source-term with argument time_index that returns an array with the values for each dof.
        :type f: function(int): Array
        :return: A function that returns the F matrix for each time index
        :rtype: function(int): Array
        """
        u,omega  = self.u,self.omega
        _ret = lambda time_index: u.multiply(f(time_index)).transpose().dot(omega) #TODO: Slow way of doing it
        return _ret


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fem4room",
    version="0.0.2.dev1",
    author="André Luiz Dalmora",
    author_email="andre.dalmora@gmail.com",
    description="Room Acoustics Simulation. Retrieve room impulse response using finite elements method.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aldalmora/fem4room",
    packages=setuptools.find_packages(),
    setup_requires=['wheel','numpy'],
    install_requires=[
        'scipy',
        'gmsh-sdk==4.4.1.post1',
        'matplotlib>=3',
        'ezdxf~=0.13.1'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Development Status :: 4 - Beta"
    ],
    python_requires='~=3.6',
    keywords="room acoustics finite elements method impulse response",
)


import gmsh
import ezdxf
import numpy as np
import matplotlib.pyplot as plt

class Mesh:
    @staticmethod
    def MeshCube(name,Lx,Ly,Lz,h,order=1):
        """Create and return an instance of a cubic mesh.

        :param name: Mesh name
        :type name: str
        :param Lx: x length of the cube
        :type Lx: float
        :param Ly: y length of the cube
        :type Ly: float
        :param Lz: z length of the cube
        :type Lz: float
        :param h: Elements size
        :type h: float
        :param order: Order of the elements, only 1 is available, defaults to 1
        :type order: int, optional
        :return: The Mesh instance.
        :rtype: FEM_3D.Mesh
        """
        m = Mesh('Mesh3D')
        m.main_dTag = m.createCube(Lx,Ly,Lz)
        m.fac.synchronize()
        dt_Boundary = m.model.getBoundary(m.main_dTag,recursive=True)
        m.model.occ.setMeshSize(dt_Boundary,h)
        m.fac.synchronize()
        m.generate(order=order)
        return m

    def __init__(self, name):
        gmsh.initialize()
        gmsh.clear()
        gmsh.model.add(name)
        self.name = name
        self.model = gmsh.model
        self.fac = self.model.occ
        self.pos = gmsh.view
        self.vertices = np.array([])
        self.tetra = np.array([])

    def calcJacobians(self):
        """Pre-calculate the determinants and inverse of the jacobian for all tetrahedrons.
        Store it in .determinants and .inverses
        """
        K = self.vertices[self.tetrahedrons]
        
        #Structure of K:
        #x = Tk(xhat) = A(xhat) + (x1) ; for Xhat 1(0,0,0), 2(1,0,0), 3(0,1,0) and 4(0,0,1):

        #A = (x2 x3 x4) - (x1)
        K = np.moveaxis(K,2,1)
        K[:,:,1] = K[:,:,1] - K[:,:,0]
        K[:,:,2] = K[:,:,2] - K[:,:,0]
        K[:,:,3] = K[:,:,3] - K[:,:,0]
        K = K[:,:,1:4]
        self.transform_matrix = K

        #Inverse of 3x3
        a = K[:,0,0];b = K[:,0,1];c = K[:,0,2]
        d = K[:,1,0];e = K[:,1,1];f = K[:,1,2]
        g = K[:,2,0];h = K[:,2,1];i = K[:,2,2]
        self.determinants = ((a*e*i) + (b*f*g) + (c*d*h) - (c*e*g) - (b*d*i) - (a*f*h))
        aux = [[e*i - f*h, c*h - b*i, b*f - c*e],
                [f*g - d*i, a*i - c*g, c*d - a*f],
                [d*h - e*g, b*g - a*h, a*e - b*d]]
        aux = np.array(aux)
        self.inverses = np.moveaxis(1/self.determinants * aux,2,0)

    def createCube(self, dx,dy,dz):
        """Add a cube to the mesh. Return the dimTag.

        :param dx: x length
        :type dx: float
        :param dy: y length
        :type dy: float
        :param dz: z length
        :type dz: float
        :return: The dimension tag(GMSH) of the cube.  (dim, Tag)
        :rtype: (int,int)
        """
        bTag = self.fac.addBox(0,0,0,dx,dy,dz)
        self.main_dTag = (3,bTag)
        return self.main_dTag

    def createSphere(self, x, y, z, radius):
        """Add a sphere to the mesh.

        :param x: Center coordinate x
        :type x: float
        :param y: Center coordinate y
        :type y: float
        :param z: Center coordinate z
        :type z: float
        :param radius: Sphere radius
        :type radius: float
        :return: The dimension tag(GMSH) of the sphere.  (dim, Tag)
        :rtype: (int,int)
        """
        sTag = self.fac.addSphere(x,y,z,radius)
        self.main_dTag = (3,sTag)
        return self.main_dTag

    def addNamedPoint(self, x, y, z, name):
        """Add a point to the mesh and a physical group to it. Return the point dimTag.

        :param x: x coordinate
        :type x: float
        :param y: y coordinate
        :type y: float
        :param z: z coordinate
        :type z: float
        :param name: Name of the point. "physical group" (GMSH)
        :type name: str
        :return: The dimension tag(GMSH) of the point.  (dim, Tag)
        :rtype: (int,int)
        """
        pTag = self.fac.addPoint(x,y,z)
        self.fac.synchronize()
        pg_point = self.model.addPhysicalGroup(0,[pTag])
        self.model.setPhysicalName(0,pg_point,name)
        return (0,pTag)

    def generate(self,order=1):
        """Generate the tetrahedrons elements of the mesh with GMSH

        :param order: Order of the elements, only 1 is available, defaults to 1
        :type order: int, optional
        """
        self.model.mesh.generate(3)
        self.model.mesh.setOrder(order)
        self.model.mesh.removeDuplicateNodes()
        self.__fillVtxTri(order)

    def readGeo(self,file):
        """Open a .GEO file for geometry descriptions 

        :param file: File path/name.
        :type file: str
        """
        gmsh.open(file)

    def readMsh(self,file):
        """Read a mshFile with GMSH. The elements must be tetrahedrons of order 1.

        :param file: File path/name
        :type file: str
        """
        gmsh.open(file)
        self.fac.synchronize()
        self.__fillVtxTri(1)

#region DXF Reading
    def __DXFReadPoint(self,e):
        """Read a point element from DXF.

        :param e: DXF point element
        :type e: DXF element
        """
        # print('Point: ', e.dxf.layer)
        point_id = self.fac.addPoint(e.dxf.location[0],e.dxf.location[1],e.dxf.location[2])
        self.dxf_point_reference = np.array([e.dxf.location[0],e.dxf.location[1],e.dxf.location[2]])
        
        self.dxf_points_ids.append(point_id)
        self.dxf_points_names.append(e.dxf.layer)

    def __DXFReadPolyMesh(self,e):
        """Read a face element from DXF.

        :param e: DXF face
        :type e: DXF element
        """
        # print('Physical Group: ', e.dxf.layer)
        for f in e.faces():
            for v in f[0:3]: #Only for three node polymeshs
                coord1=np.round(v.dxf.location[0],9)
                coord2=np.round(v.dxf.location[1],9)
                coord3=np.round(v.dxf.location[2],9)
                self.dxf_vertices.append([coord1,coord2,coord3])

            #Calculate the cross product of the triangle vectors
            p1 = np.array(self.dxf_vertices[-3])
            p2 = np.array(self.dxf_vertices[-2])
            p3 = np.array(self.dxf_vertices[-1])
            v1 = p2 - p1 
            v2 = p3 - p1
            crs =  np.dot(np.cross(v1,v2),self.dxf_point_reference)

            if crs!=0:
                self.dxf_faces_names.append(e.dxf.layer)
                # self.dxf_faces.append([len(self.dxf_vertices)-3,len(self.dxf_vertices)-2,len(self.dxf_vertices)-1])
                if crs>0:
                    self.dxf_faces.append([len(self.dxf_vertices)-3,len(self.dxf_vertices)-2,len(self.dxf_vertices)-1])
                elif crs<0:
                    self.dxf_faces.append([len(self.dxf_vertices)-1,len(self.dxf_vertices)-2,len(self.dxf_vertices)-3])
                
        
    def readDXF(self,file):
        """Read a DXF geometry description.

        :param file: File path/name
        :type file: str
        """
        self.dxf_points_ids = []
        self.dxf_points_names = []
        self.dxf_faces = []
        self.dxf_faces_names = []
        self.dxf_vertices = []

        doc = ezdxf.readfile(file)
        msp = doc.modelspace()
        for e in msp.query('POINT'): #First read the points
            if e.dxf.layer!='Layer0': #Ignore not named SketchUp points
                self.__DXFReadPoint(e)

        for e in msp.query('POLYLINE'): #Then the polymeshs
            if (e.is_poly_face_mesh):
                self.__DXFReadPolyMesh(e)

        self.dxf_vertices = np.array(self.dxf_vertices)
        self.dxf_faces = np.array(self.dxf_faces)

        #Remove duplicated points
        new_faces = np.zeros(self.dxf_faces.shape,dtype=int)
        self.dxf_vertices,inverse = np.unique(self.dxf_vertices, axis=0, return_inverse=True)
        for i in range(0,len(inverse)):
            new_faces[self.dxf_faces==i] = inverse[i]
        self.dxf_faces = new_faces

        #Add read points to the mesh. These points are used to form the surfaces
        points_surfaces_ids = []
        for v in self.dxf_vertices:
            points_surfaces_ids.append(self.fac.addPoint(v[0],v[1],v[2]))
        self.dxf_faces,idx = np.unique(self.dxf_faces,axis=0,return_index=True)
        self.dxf_faces_names = np.array(self.dxf_faces_names)[idx]

        #Stores the links between points to not recreate them when linking points to make a surface
        curves = np.zeros((len(points_surfaces_ids),len(points_surfaces_ids)))

        surface_ids = []
        for i in range(0,len(self.dxf_faces)):
            face_points_idx=self.dxf_faces[i]
            lines = []
        
            #Check if the lines of the surface already exists or create new ones
            #TODO: This part can can be simplified
            if curves[face_points_idx[0],face_points_idx[1]] != 0:
                lines.append(curves[face_points_idx[0],face_points_idx[1]])
            else:
                lId = self.fac.addLine(points_surfaces_ids[face_points_idx[0]],points_surfaces_ids[face_points_idx[1]])
                curves[face_points_idx[0],face_points_idx[1]] = lId
                curves[face_points_idx[1],face_points_idx[0]] = lId
                lines.append(lId)

            if curves[face_points_idx[1],face_points_idx[2]] != 0:
                lines.append(curves[face_points_idx[1],face_points_idx[2]])
            else:
                lId = self.fac.addLine(points_surfaces_ids[face_points_idx[1]],points_surfaces_ids[face_points_idx[2]])
                curves[face_points_idx[1],face_points_idx[2]] = lId
                curves[face_points_idx[2],face_points_idx[1]] = lId
                lines.append(lId)

            if curves[face_points_idx[0],face_points_idx[2]] != 0:
                lines.append(curves[face_points_idx[2],face_points_idx[0]])
            else:
                lId = self.fac.addLine(points_surfaces_ids[face_points_idx[0]],points_surfaces_ids[face_points_idx[2]])
                curves[face_points_idx[0],face_points_idx[2]] = lId
                curves[face_points_idx[2],face_points_idx[0]] = lId
                lines.append(lId)

            cl = self.fac.addCurveLoop(lines)
            s = self.fac.addPlaneSurface([cl])
            surface_ids.append(s)

        self.fac.synchronize()
        sl = self.fac.addSurfaceLoop(surface_ids)
        vl = self.fac.addVolume([sl])
        self.fac.healShapes()
        self.fac.synchronize()
        #Embed the defined points in DXF to the mesh. This enables the source and receiving points to be degrees of freedom in the finite elements formulation.
        self.model.mesh.embed(0,self.dxf_points_ids,3,vl)

        #Name the points and surfaces
        surface_ids = np.array(surface_ids)
        self.dxf_points_names = np.array(self.dxf_points_names)
        self.dxf_points_ids = np.array(self.dxf_points_ids)
        for n in np.unique(self.dxf_faces_names):
            idx = np.where(self.dxf_faces_names==n)
            pg = self.model.addPhysicalGroup(2,surface_ids[idx])
            self.model.setPhysicalName(2,pg,n)
        for n in np.unique(self.dxf_points_names):
            idx = np.where(self.dxf_points_names==n)
            pg = self.model.addPhysicalGroup(0,self.dxf_points_ids[idx])
            self.model.setPhysicalName(0,pg,n)
#endregion

    def __fillVtxTri(self,order):
        """Load the mesh infos(nodes and elements) from GMSH

        :param order: Order of the elements. Only 1 is available
        :type order: int
        """
        if order==1:
            tnodes = self.model.mesh.getNodesByElementType(4)
            unique, unique_indexes, unique_inverse = np.unique(tnodes[0],return_index=True,return_inverse=True)
            self.nodeTags = unique
            self.tetraTags = self.model.mesh.getElementsByType(4)[0]
            self.vertices = tnodes[1].reshape(-1,3)[unique_indexes]
            self.tetrahedrons = unique_inverse.reshape(-1,4)
        elif order==2:
            tnodes = self.model.mesh.getNodesByElementType(11)
            unique, unique_indexes, unique_inverse = np.unique(tnodes[0],return_index=True,return_inverse=True)
            self.nodeTags = unique
            self.tetraTags = self.model.mesh.getElementsByType(11)[0]
            self.vertices = tnodes[1].reshape(-1,3)[unique_indexes]
            self.tetrahedrons = unique_inverse.reshape(-1,10)

        self.vertice_group = np.zeros(len(self.vertices))

        #Points
        for n in self.model.getPhysicalGroups(0):
            nodeTags = self.model.mesh.getNodesForPhysicalGroup(n[0],n[1])[0]
            idx_nodes = np.where(nodeTags[:,None]==self.nodeTags)[1]
            self.vertice_group[idx_nodes] = n[1]

        #Surfaces
        for n in self.model.getPhysicalGroups(2):
            nodeTags = self.model.mesh.getNodesForPhysicalGroup(n[0],n[1])[0]
            idx_nodes = np.where(nodeTags[:,None]==self.nodeTags)[1]
            self.vertice_group[idx_nodes] = n[1]
        return

    def FLTKRun(self):
        """Opens the GMSH interface.
        """
        gmsh.fltk.run()